#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include "snoop.h"

typedef enum {
	AUTOSAVE,
	INVALID,
	ZOOM_FACTOR,
	ON_TOP,
	UPDATE,
	GRID,
	RED,
	GREEN,
	BLUE,
	RULER,
	HEX_MODE,
	INVERSE_Y,
	CENTER_HIGHLIGHT,
	HIGHLIGHT,
	COLORSPACE,
	FILTER,
	GRAPH,
	STATISTICS,
	GRAB_MODE,
	REFRESH_INTERVAL,
	WIDTH,
	HEIGHT,
	NOR_HISTOGRAM,
	NOR_GRADIENT,
	REF_HISTOGRAM,
	REF_GRADIENT,
	BRIGHTNESS,
	CONTRAST,
	GAMMA
} OptionField;

// useful functions for config file parsing
// get default config file name
static char* get_default_snooprc(void);
// chop leading spaces and following spaces/comments
// return the length of the result.
static int chomp(char* str);
// skip one line
static void skip_line(FILE* fin);
// get option field name
static OptionField get_option_field(const char* str);
// get option value
static int get_option_value(const char* str);
// print syntax error message
static void print_error(const char* filename, int line);

int read_config(const char* conf_file){
	char line[256];
	FILE* fin;
	int line_num = 0;
	char* filename;

	if(conf_file == NULL) // construct filename: ~/.snooprc
		filename = get_default_snooprc();
	else
		filename = (char*)conf_file;

	fin = fopen(filename, "r");
	if(fin == NULL) return 0;
	Snoop* s = global_snoop;

	while(!feof(fin)){
		line_num++;
		char *ret = fgets(line, 256, fin);
		if (!ret) {
		    fprintf(stderr, "Failed on reading config\n");
		    return 0;
		}
		if(strchr(line, '\n') == NULL){
			skip_line(fin);
			continue;
		}

		if(chomp(line) == 0)
			continue;

		OptionField field = get_option_field(line);
		if(field == GAMMA){
			// GAMMA is a double value
			int i;
			for(i = 0; line[i] != '='; i++)
				if(line[i] == '\0')
					break;

			if(line[i] == '\0'){
				print_error(filename, line_num);
				continue;
			}
			i++;
			double value;
			if(sscanf(line+i, "%lf", &value) != 1)
				print_error(filename, line_num);
			else if(value >= 0 && value <= 2)
				s->gamma = value;

			continue;
		}

		int value = get_option_value(line);
		if(value == -1){
			print_error(filename, line_num);
			continue;
		}

		switch(field){
			case AUTOSAVE:
				if(value >= 0 && value <= 1)
					s->autosave = value;
				break;

			case ZOOM_FACTOR:
				if(value >= 1 && value <= 32)
					s->zoom_factor = value;
				break;

			case UPDATE:
				if(value == 0 || value == 1)
					s->update = value;
				break;

			case ON_TOP:
				if(value == 0 || value == 1)
					s->on_top = value;
				break;

			case GRID:
				if(value == 0 || value == 1)
					s->grid = value;
				break;

			case RED:
				if(value == 0 || value == 1)
					s->red= value;
				break;

			case GREEN:
				if(value == 0 || value == 1)
					s->green = value;
				break;

			case BLUE:
				if(value == 0 || value == 1)
					s->blue = value;
				break;

			case HEX_MODE:
				if(value == 0 || value == 1)
					s->hex_mode = value;
				break;

			case INVERSE_Y:
				if(value == 0 || value == 1)
					s->inverse_y = value;
				break;

			case CENTER_HIGHLIGHT:
				if(value == 0 || value == 1)
					s->center_highlight = value;
				break;

			case HIGHLIGHT:
				if(value == 0 || value == 1)
					s->highlight = value;
				break;

			case COLORSPACE:
				if(value == 0 || value == 1 || value == 2)
					s->colorspace = value;

			case FILTER:
				if(value == 0 || value == 1)
					s->filter = value;
				break;

			case GRAPH:
				if(value >= 0 && value <=4)
					s->graph = value;
				break;

			case STATISTICS:
				if(value >= 0 && value <= 1)
					s->statistics = value;
				break;

			case GRAB_MODE:
				if(value >= 0 && value <= 1)
					s->grab_mode = value;
				break;

			case REFRESH_INTERVAL:
				if(value >= 10)
					s->refresh_interval = value;
				break;

			case RULER:
				if(value == 0 || value == 1)
					s->ruler = value;
				break;

			case WIDTH:
				s->width = value;
				break;

			case HEIGHT:
				s->height = value;
				break;

			case NOR_HISTOGRAM:
				s->nor_histogram = value;
				break;

			case NOR_GRADIENT:
				s->nor_gradient = value;
				break;

			case REF_HISTOGRAM:
				if(value >= 10 && value <= 10000)
					s->ref_histogram = value;
				break;

			case REF_GRADIENT:
				if(value >= 10 && value <= 255)
					s->ref_gradient = value;
				break;

			case BRIGHTNESS:
				s->brightness = value;
				break;

			case CONTRAST:
				s->contrast = value;
				break;

			default:
				print_error(filename, line_num);
		}
	}
	fclose(fin);
	if(conf_file == NULL)
		free(filename);

	return 1;
}

void write_config(const char* conf_file){
	char* filename;
	if(conf_file == NULL)
		filename = get_default_snooprc();
	else
		filename = (char*)conf_file;

	FILE* fout;

	Snoop* s = global_snoop;
	fout = fopen(filename, "w");
	fprintf(fout, "%s=%d\n", "autosave", s->autosave);
	fprintf(fout, "%s=%d\n", "zoom_factor", s->zoom_factor);
	fprintf(fout, "%s=%d\n", "on_top", s->on_top);
	fprintf(fout, "%s=%d\n", "update", s->update);
	fprintf(fout, "%s=%d\n", "grid", s->grid);
	fprintf(fout, "%s=%d\n", "red", s->red);
	fprintf(fout, "%s=%d\n", "green", s->green);
	fprintf(fout, "%s=%d\n", "blue", s->blue);
	fprintf(fout, "%s=%d\n", "ruler", s->ruler);
	fprintf(fout, "%s=%d\n", "hex_mode", s->hex_mode);
	fprintf(fout, "%s=%d\n", "inverse_y", s->inverse_y);
	fprintf(fout, "%s=%d\n", "center_highlight", s->center_highlight);
	fprintf(fout, "%s=%d\n", "highlight", s->highlight);
	fprintf(fout, "%s=%d\n", "colorspace", s->colorspace);
	fprintf(fout, "%s=%d\n", "filter", s->filter);
	fprintf(fout, "%s=%d\n", "graph", s->graph);
	fprintf(fout, "%s=%d\n", "statistics", s->statistics);
	fprintf(fout, "%s=%d\n", "grab_mode", s->grab_mode);
	fprintf(fout, "%s=%d\n", "refresh_interval", s->refresh_interval);
	fprintf(fout, "%s=%d\n", "width", s->width);
	fprintf(fout, "%s=%d\n", "height", s->height);
	fprintf(fout, "%s=%d\n", "histogram_normal", s->nor_histogram);
	fprintf(fout, "%s=%d\n", "histogram_reference", s->ref_histogram);
	fprintf(fout, "%s=%d\n", "gradient_normal", s->nor_gradient);
	fprintf(fout, "%s=%d\n", "gradient_reference", s->ref_gradient);
	fprintf(fout, "%s=%d\n", "brightness", s->brightness);
	fprintf(fout, "%s=%d\n", "contrast", s->contrast);
	fprintf(fout, "%s=%1.2lf\n", "gamma", s->gamma);

	fclose(fout);
	if(conf_file == NULL)
		free(filename);
}

char* get_default_snooprc(void){
	char* home = getenv("HOME");
	char* filename = (char*)malloc(strlen(home) + 10);
	sprintf(filename, "%s/.snooprc", home);
	return filename;
}

int chomp(char* str){
	int i=0, len=0;
	int j;

	while(str[i] == ' ' || str[i] == '\t') i++;
	j = i;
	while(str[j] && str[j] != '\n' && str[j] != '#'){
		j++;
		len++;
	}

	if(len == 0) return 0;
	if(i > 0) memmove(str, str+i, len);
	str[len] = '\0';
	return len;
}

void skip_line(FILE* fin){
	char str[256];
	int len;
	while(!feof(fin)){
		char *s = fgets(str, 256, fin);
		if (!s) continue;
		len = strlen(str);
		if(str[len-1] == '\n')
			return;
	}
}

OptionField get_option_field(const char* str){
	// handmade tree search
	switch(str[0]){
		case 'a':
			if(  !strncmp(str, "autosave", 8) )
				return AUTOSAVE;
			break;

		case 'b':
			if ( !strncmp(str, "blue", 4) )
				return BLUE;
			else if( !strncmp(str, "brightness", 10) )
				return BRIGHTNESS;
			break;
		case 'c':
			if( !strncmp(str, "center_highlight", 16) )
				return CENTER_HIGHLIGHT;
			else if( !strncmp(str, "contrast", 8) )
				return CONTRAST;
			else if ( !strncmp(str, "colorspace", 10) )
				return COLORSPACE;
			break;

		case 'f':
			if( !strncmp(str, "filter", 6) )
				return FILTER;
			break;

		case 'g':
			if( str[2] == 'a' ){
				if( !strncmp(str, "graph", 5) )
					return GRAPH;
				else if( !strncmp(str, "grab_mode", 9) )
					return GRAB_MODE;
				else if( !strncmp(str, "gradient_normal", 15) )
					return NOR_GRADIENT;
				else if( !strncmp(str, "gradient_reference", 18) )
					return REF_GRADIENT;
			}else if( !strncmp(str, "grid", 4) )
				return GRID;
			else if( !strncmp(str, "green", 5) )
				return GREEN;
			else if( !strncmp(str, "gamma", 5))
				return GAMMA;
			break;

		case 'h':
			if( str[2] == 'x' && !strncmp(str, "hex_mode", 8) )
				return HEX_MODE;
			else if( !strncmp(str, "height", 6) )
				return HEIGHT;
			else if( !strncmp(str, "highlight", 9) )
				return HIGHLIGHT;
			else if( !strncmp(str, "histogram_normal", 16) )
				return NOR_HISTOGRAM;
			else if( !strncmp(str, "histogram_reference", 19) )
				return REF_HISTOGRAM;
			break;

		case 'i':
			if( !strncmp(str, "inverse_y", 9) )
				return INVERSE_Y;
			break;

		case 'o':
			if( !strncmp(str, "on_top", 6) )
				return ON_TOP;
			break;

		case 'r':
			if( !strncmp(str, "refresh_interval", 16) )
				return REFRESH_INTERVAL;
			else if( !strncmp(str, "ruler", 5) )
				return RULER;
			else if ( !strncmp(str, "red", 3) )
				return RED;
			break;

		case 's':
			if( !strncmp(str, "statistics", 10))
				return STATISTICS;
			break;

		case 'u':
			if( !strncmp(str, "update", 6) )
				return UPDATE;
			break;

		case 'w':
			if( !strncmp(str, "width", 5) )
				return WIDTH;
			break;

		case 'z':
			if( !strncmp(str, "zoom_factor", 11) )
				return ZOOM_FACTOR;
			break;

		default:
			return INVALID;
	}
	return INVALID;
}

int get_option_value(const char* str){
	int value;
	int i=0;

	while(str[i] && str[i] != '=') i++;

	if(str[i] == '\0') return 0;

	i++;
	if(sscanf(str+i, "%d", &value) == 0)
		return -1;
	else
		return value;
}

void print_error(const char* filename, int line){
	fprintf(stderr, "Warning: syntax error in config file %s line %d\n", filename, line);
}
